import base64
import binascii
import json
import os
import struct
import time
import requests
import urllib
from base64 import b64encode
from datetime import datetime
from hashlib import md5

import axolotl_curve25519 as curve
import Crypto.Cipher.PKCS1_OAEP as rsaenc
import httpx
import xxhash
from Crypto.Cipher import AES
from Crypto.PublicKey import RSA
from Crypto.Util.Padding import pad, unpad
from thrift.transport.TTransport import TMemoryBuffer

from .services.thrift.ttypes import TalkException
from .exceptions import LineServiceException
from .serializers.DummyProtocol import (
    DummyProtocol,
    DummyProtocolData,
    DummyProtocolSerializer,
    DummyThrift,
)

from .services.thrift import *
from .timeline import Timeline
from .services.thrift.ap.TCompactProtocol import TCompactProtocol as tcProtocol


class Models(object):
    def __init__(self, savePath):
        self.savePath = savePath or os.path.dirname(os.path.realpath(__file__))
        self.lcsStart = "0005"
        self.le = "18"
        self.PUBLIC_KEY = "-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA0LRokSkGDo8G5ObFfyKiIdPAU5iOpj+UT+A3AcDxLuePyDt8IVp9HpOsJlf8uVk3Wr9fs+8y7cnF3WiY6Ro526hy3fbWR4HiD0FaIRCOTbgRlsoGNC2rthp2uxYad5up78krSDXNKBab8t1PteCmOq84TpDCRmainaZQN9QxzaSvYWUICVv27Kk97y2j3LS3H64NCqjS88XacAieivELfMr6rT2GutRshKeNSZOUR3YROV4THa77USBQwRI7ZZTe6GUFazpocTN58QY8jFYODzfhdyoiym6rXJNNnUKatiSC/hmzdpX8/h4Y98KaGAZaatLAgPMRCe582q4JwHg7rwIDAQAB\n-----END PUBLIC KEY-----"
        self.key = RSA.importKey(self.PUBLIC_KEY)
        self.encryptKey = os.urandom(16)
        self.IV = bytes(
            [78, 9, 72, 62, 56, 245, 255, 114, 128, 18, 123, 158, 251, 92, 45, 51]
        )
        self.cipher = AES.new(self.encryptKey, AES.MODE_CBC, iv=self.IV)
        self.d_cipher = AES.new(self.encryptKey, AES.MODE_CBC, iv=self.IV)
        self.encEncKey()
        # self.initWithBiz()
        # self.initWithAndroid(4)

        # Init 3rd Models
        from .dyher.connManager import ConnManager

        self.legyPushers = ConnManager(self)

    def getSavePath(self, dirname: str):
        savePath = os.path.join(self.savePath, dirname)
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        return savePath

    def log(self, text, debugOnly=False):
        if debugOnly and not self.isDebug:
            return
        print(f"[{datetime.now():%Y-%m-%d %H:%M:%S}] {text}")

    def genOBSParams(self, newList, returnAs="json", ext="jpg"):
        oldList = {"name": f"CHRLINE-{int(time.time())}.{ext}", "ver": "1.0"}
        if returnAs not in ["json", "b64", "default"]:
            raise Exception("Invalid parameter returnAs")
        oldList.update(newList)
        if "range" in oldList:
            new_range = "bytes 0-%s\/%s" % (
                str(oldList["range"] - 1),
                str(oldList["range"]),
            )
            oldList.update({"range": new_range})
        if returnAs == "json":
            oldList = json.dumps(oldList)
            return oldList
        elif returnAs == "b64":
            oldList = json.dumps(oldList)
            return b64encode(oldList.encode("utf-8"))
        elif returnAs == "default":
            return oldList

    def checkNextToken(self, log4NotDebug: bool = True):
        savePath = self.getSavePath(".tokens")
        fn = md5(self.authToken.encode()).hexdigest()
        old = self.authToken
        if os.path.exists(savePath + f"/{fn}"):
            self.authToken = open(savePath + f"/{fn}", "r").read()
            if old != self.authToken:
                self.log(f"New Token: {self.authToken}", not log4NotDebug)
                self.checkNextToken(log4NotDebug)
        return self.authToken

    def handleNextToken(self, newToken):
        savePath = self.getSavePath(".tokens")
        fn = md5(self.authToken.encode()).hexdigest()
        open(savePath + f"/{fn}", "w").write(newToken)
        self.authToken = newToken
        self.log(f"New Token: {newToken}")
        Timeline.__init__(self)

    def tryRefreshToken(self):
        """Try to refresh token."""
        refreshToken = self.getCacheData(".refreshToken", self.authToken)
        self.log(f"try to refresh access token... {refreshToken}")
        if refreshToken is not None:
            RATR = self.refreshAccessToken(refreshToken)
            token = self.checkAndGetValue(RATR, "accessToken", 1)
            # refreshToken = self.checkAndGetValue(RATR, 'refreshToken', 5)
            self.handleNextToken(token)
            self.saveCacheData(".refreshToken", token, refreshToken)
        else:
            raise ValueError("RefreshToken missing.")
        return True

    def getCustomData(self):
        savePath = self.getSavePath(".data")
        fn = md5(self.customDataId.encode()).hexdigest()
        if os.path.exists(savePath + f"/{fn}"):
            self.custom_data = json.loads(open(savePath + f"/{fn}", "r").read())
        self.log(f"Loading Custom Data: {fn}")
        return True

    def saveCustomData(self):
        savePath = self.getSavePath(".data")
        fn = md5(self.customDataId.encode()).hexdigest()
        open(savePath + f"/{fn}", "w").write(json.dumps(self.custom_data))
        return True

    def getSqrCert(self):
        savePath = self.getSavePath(".data")
        fn = "cert.pem"
        if os.path.exists(savePath + f"/{fn}"):
            return open(savePath + f"/{fn}", "r").read()
        return None

    def saveSqrCert(self, cert):
        savePath = self.getSavePath(".data")
        fn = "cert.pem"
        open(savePath + f"/{fn}", "w").write(cert)
        return True

    def getEmailCert(self, email):
        savePath = self.getSavePath(".data")
        fn = f"{email}.crt"
        if os.path.exists(savePath + f"/{fn}"):
            return open(savePath + f"/{fn}", "r").read()
        return None

    def saveEmailCert(self, email, cert):
        savePath = self.getSavePath(".data")
        fn = f"{email}.crt"
        open(savePath + f"/{fn}", "w").write(cert)
        return True

    def initWithAndroid(self, ver=7):
        if ver == 1:
            self.lcsStart = "0001"
            self.le = "1"
            self.PUBLIC_KEY = "-----BEGIN PUBLIC KEY-----\nMIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQCZAAoZNRwIlLUXaUgrgYi8bAYq\nQeFVtXvCNIEm+F4/jAyTU3YoDwmoLaKQ6itGOonykGtwy2k/3BeWefL/q5eUGjVG\nBEa1vBeUNEb4IFU8n9WK3N/GIIPuD6ZiusB+U1FPg/NaEiVX8ldmEQJgmuG1hykk\n2dU3oy7O1M+Kwl1lJQIDAQAB\n-----END PUBLIC KEY-----"
        elif ver == 4:
            self.lcsStart = "0004"
            self.le = "7"  # LegyEncHelper.cpp::decryptStream -> legy xle value
            self.PUBLIC_KEY = "-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAwpAwTVluR1Z++tVzxtOD\nr7XxSv6oqrwvj/8c8SkfFsS8zM7CvIT8j+x+6Qs1JjNRDtYjAwPKO3tO+qOAdA+8\n7FHpx0THDJIi4VYxSZ2uDh0U8Luxh02whwM8gPbPQNN3sEd5++kJ3cCh5eeAIiUd\nDrwPhHzxO8swpBRdxJB/pzibEqpG2U2764JlPscN9D896qmBN6CBRKpXk/MmUDAI\n4xg+uQk/ykn3SNXJSgQwI1UD9KuiR+X9tbJlKRMN5JpUrSuEwRPQQDMaWpSIdCJM\noFqJLNwt9b1RR/JEB01Eup+3QCub20/CObCmHZY6G26KTDHLoTRZ1xzymdYhdJ43\nCwIDAQAB\n-----END PUBLIC KEY-----"
        elif ver == 6:
            self.lcsStart = "0007"
            self.le = "7"
            self.PUBLIC_KEY = "-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA5ABzJbexh+HH1+RzVTH4\nDFj8b/42vRqUp0NWLIBAgi5+bAgeJYzyVBI7Pk6YkQnd44HPvUvFMF3V3TocRXLP\nZV/NSawgcAh+VrXe3VIlruCOe14qrd/ZMpTRTxtBJ5aRpVhTsnGpZtGggPYnPh4c\n6V/R7Wxymj4SBZ1Ipsa7ZI83z/XIPFXT38qTJN3UAW5YhjQon1eNbwaxALVajuvI\npUE52aIBi05gE/V0HEoOUjfOg1V6RHFbcchTgmcze6Vbye+7kmdsIboDXnNpm/fJ\nuItub+iwLKSWf9OPc/bYpU4YVBxZXzmSCXFMLeCe2i5wJeMg4iG8NpVpwVv2W+Hb\nQQIDAQAB\n-----END PUBLIC KEY-----"
        elif ver == 7:
            self.lcsStart = "0008"
            self.le = "7"
            self.PUBLIC_KEY = "-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAsMC6HAYeMq4R59e2yRw6\nW1OWT2t9aepiAp4fbSCXzRj7A29BOAFAvKlzAub4oxN13Nt8dbcB+ICAufyDnN5N\nd3+vXgDxEXZ/sx2/wuFbC3B3evSNKR4hKcs80suRs8aL6EeWi+bAU2oYIc78Bbqh\nNzx0WCzZSJbMBFw1VlsU/HQ/XdiUufopl5QSa0S246XXmwJmmXRO0v7bNvrxaNV0\ncbviGkOvTlBt1+RerIFHMTw3SwLDnCOolTz3CuE5V2OrPZCmC0nlmPRzwUfxoxxs\n/6qFdpZNoORH/s5mQenSyqPkmH8TBOlHJWPH3eN1k6aZIlK5S54mcUb/oNRRq9wD\n1wIDAQAB\n-----END PUBLIC KEY-----"
        self.key = RSA.importKey(self.PUBLIC_KEY)
        self.IV = bytes(
            [78, 9, 72, 62, 56, 245, 255, 114, 128, 18, 123, 158, 251, 92, 45, 51]
        )
        self.encryptKey = b"DearSakura+2021/"
        self.encEncKey()

    def encHeaders(self, headers):
        t = headers.keys()
        data = []
        self.mFhrnmxnNF(len(t), data)
        for i in t:
            self.mFhrnmxnNF(len(i), data)
            self.wYEpEYldst(i, data)
            self.mFhrnmxnNF(len(headers[i]), data)
            self.wYEpEYldst(headers[i], data)
        o = len(data)
        data = [255 & o] + data
        data = [255 & o >> 8] + data
        return data

    def decHeaders(self, data):
        headers = {}
        tbin = self.TBinaryProtocol(self)
        tbin.data = data
        dataLen = tbin.readI16() + 2
        headerLen = tbin.readI16()
        for _ in range(headerLen):
            size = tbin.readI16()
            _k = tbin.y(size).decode()
            size = tbin.readI16()
            _v = tbin.y(size).decode()
            headers[_k] = _v
        return headers, data[dataLen:]

    def encEncKey(self):
        # heh
        a = rsaenc.new(self.key)
        self._encryptKey = (
            self.lcsStart + b64encode(a.encrypt(self.encryptKey)).decode()
        )

    def encData(self, data):
        _data = AES.new(self.encryptKey, AES.MODE_CBC, iv=self.IV).encrypt(
            pad(data, AES.block_size)
        )
        debug = []
        return _data

    def decData(self, data):
        data = pad(data, AES.block_size)
        _data = AES.new(self.encryptKey, AES.MODE_CBC, iv=self.IV).decrypt(data)[:-16]
        _data = unpad(_data, AES.block_size)
        i = 1
        data = self.yVdzCLDwMN(_data, i)
        i = 3
        return _data

    def mFhrnmxnNF(self, t, e):
        i = 65536
        if t < -1 * 32768 or t >= i:
            raise Exception(t + " is incorrect for i16.")
        e.append(255 & t >> 8)
        e.append(255 & t)

    def wYEpEYldst(self, t, e):
        for i in range(len(t)):
            e.append(ord(t[i]))

    def xZVpUuXFru(t):
        if 8 == len(t):
            return t
        e = ""
        i = 0
        n = 8 - len(t)
        while i < n:
            e += "0"
            i += 1
        return e + t

    def pmAWhahfKx(self, t):
        e = []
        i = 0
        n = len(t)
        while i < n:
            _i = 0
            try:
                _i = int(t[i : i + 2], 16)
            except:
                _i = 16
            e.append(_i)
            i += 2
        return e

    def XQqwlHlXKK(self, e, i):
        r = []
        for o in range(16):
            r.append(92 ^ e[o])
        n = xxhash.xxh32(b"", seed=0)
        s = xxhash.xxh32(b"", seed=0)
        n.update(bytes(r))
        for o in range(16):
            r[o] ^= 106
        s.update(bytes(r))
        s.update(i)
        a = s.hexdigest()  # is b8a7c677?
        n.update(bytes(self.pmAWhahfKx(a)))
        c = n.hexdigest()  # is 3f97d2f6?
        d = self.pmAWhahfKx(c)
        return bytes(d)

    def yVdzCLDwMN(self, d, i):
        return (255 & self.xnEmbaRWhy(d, i)) << 8 | 255 & self.xnEmbaRWhy(d, i + 1)

    def xnEmbaRWhy(self, d, i):
        t = d[i]
        if t > 127:
            t = 0 - (t - 1 ^ 255)
        return t

    def generateDummyProtocol(
        self, name: str, params: list, _type: int
    ) -> DummyProtocolSerializer:
        return DummyProtocolSerializer(self, name, params, _type)

    def generateDummyProtocol2(
        self, params: DummyProtocol, proto: int = 3, fixSuccessHeaders: bool = False
    ):
        newParams = []
        d = params.data
        if d is not None:
            newParams.append(thrift2dummy(d))
        return bytes(self.generateDummyProtocolField(newParams, proto) + [0])

    def generateDummyProtocolField(self, params, proto):
        isCompact = False
        data = []
        tcp = self.TCompactProtocol(self)
        for param in params:
            # [10, 2, revision]
            _type = param[0]
            _id = param[1]
            _data = param[2]
            if _data is None:
                continue
            if _type == 13:
                if _data[2] is None:
                    continue
            elif _type in [14, 15]:
                if _data[1] is None:
                    continue
            if proto == 3:
                data += [_type, 0, _id]
                isCompact = False
            elif proto == 4:
                if _type == 2:
                    data += tcp.getFieldHeader(0x01 if _data else 0x02, _id)
                    continue
                data += tcp.getFieldHeader(tcp.CTYPES[_type], _id)
                isCompact = True
            data += self.generateDummyProtocolData(_data, _type, isCompact)
        return data

    def generateDummyProtocolData(self, _data, ttype, isCompact=False):
        data = []
        tbp = self.TBinaryProtocol(self)
        tcp = self.TCompactProtocol(self)
        proto = 4 if isCompact else 3
        if ttype == 2:
            if isCompact:
                # CONTAINER_WRITE
                data += tcp.writeByte(0x01 if _data else 0x02)
            else:
                data += [1] if _data is True else [0]
        elif ttype == 3:
            if isCompact:
                data += tcp.writeByte(_data)
            else:
                data += tbp.writeByte(_data)
        elif ttype == 4:
            data = self.getFloatBytes(_data, isCompact=isCompact)
        elif ttype == 8:
            data += self.getIntBytes(_data, isCompact=isCompact)
        elif ttype == 10:
            data += self.getIntBytes(_data, 8, isCompact=isCompact)
        elif ttype == 11:
            data += self.getStringBytes(_data, isCompact=isCompact)
        elif ttype == 12:
            if isinstance(_data, DummyProtocolData):
                _data = thrift2dummy(_data)
            data += self.generateDummyProtocolField(_data, proto) + [0]
        elif ttype == 13:
            _ktype = _data[0]
            _vtype = _data[1]
            _vdata = _data[2]
            if isCompact:
                data += tcp.writeMapBegin(_ktype, _vtype, len(_vdata))
            else:
                data += [_ktype, _vtype] + self.getIntBytes(
                    len(_vdata), isCompact=isCompact
                )
            for vd in _vdata:
                data += self.generateDummyProtocolData(vd, _ktype, isCompact)
                data += self.generateDummyProtocolData(_vdata[vd], _vtype, isCompact)
        elif ttype == 14 or ttype == 15:
            # [11, targetUserMids]
            _vtype = _data[0]
            _vdata = _data[1]
            if isCompact:
                data += tcp.writeCollectionBegin(_vtype, len(_vdata))
            else:
                data += [_vtype] + self.getIntBytes(len(_vdata), isCompact=isCompact)
            for vd in _vdata:
                data += self.generateDummyProtocolData(vd, _vtype, isCompact)
        else:
            raise Exception(f"[generateDummyProtocolData] not support type: {type}")
        return data

    def postPackDataAndGetUnpackRespData(
        self,
        path: str,
        bdata: bytes,
        ttype: int = 3,
        encType: int = None,
        headers: dict = None,
        access_token: str = None,
        baseException: dict = None,
        readWith: str = None,
        conn: any = None,
        files: dict = None,
        expectedRespCode: list = None,
        timeout: int = None,
    ):
        if expectedRespCode is None:
            expectedRespCode = [200]
        if headers is None:
            headers = self.server.Headers.copy()
        if access_token is None:
            access_token = self.authToken
        ptype = "TBINARY" if ttype == 3 else "TCOMPACT"
        if ttype in [1, 2, 3, 4, 5]:
            headers["content-type"] = "application/x-thrift; protocol=" + ptype
            headers["accept"] = "application/x-thrift"
            if isinstance(bdata, DummyProtocolSerializer):
                # Import DummyProtocolSerializer in v2.5.3,
                # it can be change the protocol type.
                if self.force_tmore:
                    # Force TMoreCompact
                    if path in [self.LINE_NORMAL_ENDPOINT, "/S4"]:
                        path = "/S5"
                        ttype = 5
                bdata.protocol = ttype
        headers["x-lal"] = self.LINE_LANGUAGE
        if encType is None:
            encType = self.encType

        # 2022/08/24 PATCH
        if self.DEVICE_TYPE == "CHROMEOS":
            headers[
                "origin"
            ] = "chrome-extension://CHRLINE-v2.5.0-rc-will-not-be-released"

        data = bdata
        if type(data) in [DummyProtocolSerializer, list]:
            data = bdata = bytes(bdata)
        elif not isinstance(bdata, bytes):
            bdata = str(bdata).encode()
        self.log(f"----------------- START POST", True)
        self.log(
            f"--> POST {path} {f'({self.LINE_ENCRYPTION_ENDPOINT})' if encType == 1 else ''}",
            True,
        )
        self.log(
            f"--> {bdata.hex()}",
            True,
        )
        if encType == 0:
            if conn is None:
                conn = self.req_h2
            if "x-le" in headers:
                del headers["x-le"]
                del headers["x-lcs"]
            if access_token is not None:
                headers["X-Line-Access"] = access_token
            self.log(f"--> Headers: {headers}", True)
            res = doLoopReq(
                conn.post,
                {
                    "url": self.LINE_GW_HOST_DOMAIN + path,
                    "data": data,
                    "headers": headers,
                    "files": files,
                    "timeout": timeout,
                },
            )
            data = res.content
        elif encType == 1:
            if conn is None:
                conn = self.req
            if access_token is not None:
                _headers = {"x-lt": access_token, "x-lpqs": path}
            else:
                _headers = {"x-lpqs": path}
            a = self.encHeaders(_headers)
            _data = bytes(a) + bytes(bdata)
            fix_bytes = False
            if (int(self.le) & 4) == 4:
                _data = bytes([int(self.le)]) + _data
                fix_bytes = True
            if (int(self.le) & 2) != 2:
                data = self.encData(_data)
            else:
                data = self.encData(_data)
                data += self.XQqwlHlXKK(self.encryptKey, data)
            headers["accept-encoding"] = "gzip, deflate"
            self.log(f"--> Headers: {headers} ({_headers})", True)
            res = doLoopReq(
                conn.post,
                {
                    "url": self.LINE_GF_HOST_DOMAIN + self.LINE_ENCRYPTION_ENDPOINT,
                    "data": data,
                    "files": files,
                    "headers": headers,
                    "timeout": timeout,
                },
            )
            if res.content:
                data = self.decData(res.content)
            else:
                data = res.content
            if fix_bytes:
                data = data[1:]
        else:
            raise Exception(f"Unknown encType: {encType}")
        self.log(f"<--  {res.status_code}", True)
        self.log(f"{data.hex()}", True)
        if res.status_code in expectedRespCode:
            if (
                res.headers.get("x-lc") is not None
                and int(res.headers["x-lc"]) not in expectedRespCode
            ):
                raise Exception(f'Invalid response code: {res.headers["x-lc"]}')
            if encType == 1:
                respHeaders, data = self.decHeaders(data)
            else:
                respHeaders = {}
            respHeaders.update(res.headers)
            self.log(f"RespHraders: {respHeaders}", True)
            for _rh in respHeaders:
                _rht = str(_rh)
                if _rht == "x-line-next-access":
                    self.handleNextToken(respHeaders[_rh])
                elif _rht.startswith("x-"):
                    if _rht not in [
                        "x-ls",
                        "x-lc",
                        "x-le",
                        "x-line-http",
                        "x-lcr",
                        "x-lts",
                        "x-obs-oid",
                        "x-obs-hash",
                        "x-obs-debug-object-location",
                        "x-obs-debug-id",
                        "x-obs-content-type",
                        # "x-line-next-access-max-age",
                    ]:
                        self.log(
                            f"[HTTP] unhandled header:  {_rht} => {respHeaders[_rh]}"
                        )
            # x-line-access-refresh-required
            conn_res = res
            res = None
            if ttype == -7:
                # COMPACT
                # TODO: ADD DECODER
                compact = self.TCompactProtocol(self)
                _type = compact.readByte(data)
                data = data[1:]
                if _type == 1:
                    _seq, _offset = compact.readI32(data, True)
                    data = data[_offset:]
                    _msgId, _offset = compact.readI32(data, True)
                    data = data[_offset:]
                    _ts, _offset = compact.readI32(data, True)
                    data = data[_offset:]
                    compact.res = {"_seq": _seq, "messageId": _msgId, "time": _ts}
                    res = compact
                elif _type == 2:
                    raise LineServiceException({"code": compact.readI32(data)})
            elif ttype == -3:
                # JSON RAW
                return json.loads(data)
            elif ttype == -1:
                # CONTENT RAW
                return data
            elif ttype == 0:
                # RESP
                return conn_res
            else:
                # THRIFT
                try:
                    if ttype == 3:
                        res = self.TBinaryProtocol(
                            self, data, baseException=baseException
                        )
                    elif ttype == 4:
                        res = self.TCompactProtocol(
                            self, data, baseException=baseException
                        )
                    elif ttype == 5:
                        tmore = self.TMoreCompactProtocol(
                            self, data, baseException=baseException
                        )
                        res = tmore
                    else:
                        raise ValueError(f"Unknown ThriftType: {ttype}")
                    if (
                        self.use_thrift
                        and getattr(res, "dummyProtocol", None) is not None
                    ):
                        res = self.serializeDummyProtocolToThrift(
                            res.dummyProtocol, baseException, readWith
                        )
                    else:
                        res = res.res
                except Exception as e:
                    _err = {
                        "code": None,
                        "message": None,
                        "metadata": None,
                        "raw": None,
                    }
                    if isinstance(e, LineServiceException):
                        _err["code"] = e.code
                        _err["message"] = e.message
                        _err["metadata"] = e.metadata
                    elif isinstance(e, TalkException):
                        _err["code"] = e.code
                        _err["message"] = e.reason
                        _err["metadata"] = e.parameterMap
                        _err["raw"] = e
                    else:
                        # non handler
                        raise e
                    res = {"error": _err}
            if type(res) == dict and "error" in res:
                # idk why it got int on sometime
                resMsg = str(res["error"]["message"])
                logOutList = [
                    "EXPIRED",
                    "REVOKE",
                    "LOG_OUT",
                    "AUTHENTICATION_DIVESTED_BY_OTHER_DEVICE",
                    "DEVICE_LOSE",
                    "IDENTIFY_MODIFIED",
                    "V3_TOKEN_CLIENT_LOGGED_OUT",
                    "DELETED_ACCOUNT",
                ]
                if resMsg is not None and (
                    resMsg in logOutList or resMsg.startswith("suspended")
                ):
                    self.is_login = False
                    self.log(f"LOGIN OUT: {resMsg}")
                elif res["error"]["code"] == 119:
                    if self.tryRefreshToken():
                        return self.postPackDataAndGetUnpackRespData(
                            path, bdata, ttype, encType, headers
                        )
                    self.log(f"LOGIN OUT: {resMsg}")
                print(res)
                raise LineServiceException(res["error"])
            self.log(f"Result: {res}", True)
            self.log(f"----------------- END POST", True)
            return res
        elif res.status_code in [400, 401, 403]:
            self.is_login = False
        elif res.status_code == 410:
            return None
        raise Exception(
            f"Invalid response status code: {res.status_code}, Headers: {res.headers}"
        )

    def getCurrReqId(self):
        self._msgSeq = 0
        if "_reqseq" in self.custom_data:
            self._msgSeq = self.custom_data["_reqseq"]
        self._msgSeq += 1
        self.custom_data["_reqseq"] = self._msgSeq
        self.saveCustomData()
        return self._msgSeq

    def getIntBytes(self, i, j=4, isCompact=False):
        i = int(i)
        if isCompact:
            _compact = self.TCompactProtocol(self)
            a = _compact.makeZigZag(i, 32 if j**2 == 16 else 64)
            b = _compact.writeVarint(a)
            return b
        if j**2 == 16:
            res = struct.pack("!i", i)
        else:
            res = struct.pack("!q", i)
        return list(res)

    def getStringBytes(self, text, isCompact=False):
        if text is None:
            text = ""
        if type(text) == bytes:
            pass
        else:
            text = str(text).encode()
        if isCompact:
            _compact = self.TCompactProtocol(self)
            sqrd = _compact.writeVarint(len(text))
        else:
            sqrd = self.getIntBytes(len(text))
        for value in text:
            sqrd.append(value)
        return sqrd

    def getFloatBytes(self, val, isCompact=False):
        res = []
        _t = "!d"
        if isCompact:
            _t = "<d"
        for value in struct.pack(_t, val):
            res.append(value)
        return res

    def getMagicStringBytes(self, val, rev=False):
        res = []
        if rev:
            res = binascii.b2a_hex(val)
        else:
            if len(val) == 32:
                for ii in range(16):
                    iii = ii * 2
                    i = iii + 1
                    mgc = (int(val[iii], 16) << 4) + int(val[i], 16)
                    res.append(mgc)
            else:
                raise ValueError(
                    f"getMagicStringBytes() expected 32, but got {len(val)}"
                )
        return res

    def createSqrSecret(self, base64Only=False):
        private_key = curve.generatePrivateKey(os.urandom(32))
        public_key = curve.generatePublicKey(private_key)
        secret = urllib.parse.quote(b64encode(public_key).decode())
        version = 1
        if base64Only:
            return [private_key, b64encode(public_key).decode()]
        return [private_key, f"?secret={secret}&e2eeVersion={version}"]

    def getE2EESelfKeyData(self, mid):
        savePath = self.getSavePath(".e2eeKeys")
        fn = f"{mid}.json"
        if os.path.exists(savePath + f"/{fn}"):
            return json.loads(open(savePath + f"/{fn}", "r").read())
        keys = self.getE2EEPublicKeys()
        for key in keys:
            keyId = self.checkAndGetValue(key, "keyId", 2)
            _keyData = self.getE2EESelfKeyDataByKeyId(keyId)
            if _keyData is not None:
                return _keyData
        raise Exception("E2EE Key has not been saved, try register or use SQR Login")

    def getE2EESelfKeyDataByKeyId(self, keyId):
        savePath = self.getSavePath(".e2eeKeys")
        fn = f"key_{keyId}.json"
        if os.path.exists(savePath + f"/{fn}"):
            return json.loads(open(savePath + f"/{fn}", "r").read())
        return None

    def saveE2EESelfKeyData(self, mid, pubK, privK, kI, e2eeVersion):
        savePath = self.getSavePath(".e2eeKeys")
        fn = f"{mid}.json"
        fn2 = f"key_{kI}.json"
        data = json.dumps(
            {
                "keyId": kI,
                "privKey": b64encode(privK).decode(),
                "pubKey": b64encode(pubK).decode(),
                "e2eeVersion": e2eeVersion,
            }
        )
        if mid is not None:
            open(savePath + f"/{fn}", "w").write(data)
        open(savePath + f"/{fn2}", "w").write(data)
        return True

    def registerE2EESelfKey(self, privK: bytes = None):
        if privK is None:
            privK = curve.generatePrivateKey(os.urandom(32))
        if len(privK) != 32:
            raise ValueError("Invalid private key.")
        pubK = curve.generatePublicKey(privK)
        EPK = self.registerE2EEPublicKey(1, None, pubK, 0)
        keyId = self.checkAndGetValue(EPK, "keyId", 2)
        return self.saveE2EESelfKeyData(self.mid, pubK, privK, keyId, 1)

    def getCacheData(self, cT, cN, needHash=True, pathOnly=False):
        savePath = self.getSavePath(cT)
        fn = f"{cN}"
        if needHash:
            fn = md5(cN.encode()).hexdigest()
        if pathOnly:
            return savePath + f"/{fn}"
        if os.path.exists(savePath + f"/{fn}"):
            return open(savePath + f"/{fn}", "r").read()
        return None

    def saveCacheData(self, cT, cN, cD, needHash=True):
        savePath = self.getSavePath(cT)
        fn = f"{cN}"
        if needHash:
            fn = md5(cN.encode()).hexdigest()
        data = cD
        open(savePath + f"/{fn}", "w").write(data)
        return True

    def decodeE2EEKeyV1(self, data: dict, secret: bytes, mid: str = None):
        if "encryptedKeyChain" in data:
            print("Try to decode E2EE Key")
            encryptedKeyChain = base64.b64decode(data["encryptedKeyChain"])
            # hashKeyChain = data['hashKeyChain']
            keyId = data["keyId"]
            publicKey = base64.b64decode(data["publicKey"])
            e2eeVersion = data["e2eeVersion"]
            e2eeKey = self.decryptKeyChain(publicKey, secret, encryptedKeyChain)
            print(f"E2EE Priv Key: {e2eeKey[0]}")
            print(f"E2EE Pub Key: {e2eeKey[1]}")
            print(f"keyId: {keyId}")
            print(f"e2eeVersion: {e2eeVersion}")
            self.saveE2EESelfKeyData(mid, e2eeKey[1], e2eeKey[0], keyId, e2eeVersion)
            return {
                "keyId": keyId,
                "privKey": e2eeKey[0],
                "pubKey": e2eeKey[1],
                "e2eeVersion": e2eeVersion,
            }

    def tryReadThriftContainerStruct(self, data, id=0, get_data_len=False):
        _data = {}
        _dec = self.TCompactProtocol(self)
        ftype = data[0] & 15
        fid = (data[0] >> 4) + id
        offset = 1
        nextPos = 0
        if ftype == 0:
            _data = None
        elif ftype == 4:
            _data[fid] = _dec.readDouble(data[offset:])
            nextPos += 8
        elif ftype == 5:
            (_data[fid], nextPos) = _dec.readI32(data[offset:], True)
            nextPos += 1
        elif ftype == 6:
            (_data[fid], nextPos) = _dec.readI64(data[offset:], True)
            nextPos += 1
        elif ftype == 8:
            (_data[fid], nextPos) = _dec.readBinary(data[offset:])
        elif ftype == 9 or ftype == 10:
            (vtype, vsize, vlen) = _dec.readCollectionBegin(data[offset:])
            offset += vlen
            _data[fid] = []
            _nextPos = 0
            for i in range(vsize):
                if vtype == 12:
                    _aaa, _bbb = self.tryReadThriftContainerStruct(
                        data[offset:], get_data_len=True
                    )
                    _data[fid].append(_aaa)
                    offset += _bbb + 1
        else:
            print(f"[tryReadThriftContainerStruct]不支援Type: {ftype} => ID: {fid}")
        if nextPos > 0:
            data = data[nextPos:]
            c = self.tryReadThriftContainerStruct(data, id=fid, get_data_len=True)
            if c[0] is not None:
                _data.update(c[0])
            nextPos += c[1]
        if get_data_len:
            return [_data, nextPos]
        return _data

    def serializeDummyProtocolToThrift(
        self, data: DummyProtocol, baseException: dict = None, readWith: str = None
    ):
        if baseException is None:
            baseException = {}
        if readWith is not None:
            new1 = self.generateDummyProtocol2(data, 4, fixSuccessHeaders=True)
            try:
                a = eval(f"{readWith}_result")
                a = a()
            except AttributeError:
                a = None
            except NameError:
                a = None
            if a is not None:
                e = TMemoryBuffer(new1)
                f = tcProtocol(e)
                a.read(f)
                if getattr(a, "success", None) is not None:
                    return a.success
                if getattr(a, "e", None) is not None:
                    raise a.e
                return None

        def _gen():
            return DummyThrift()

        def _get(a):
            return a.data if isinstance(a, DummyProtocolData) else a

        def _genFunc(a: DummyProtocolData, b, f):
            def __gen(a: DummyProtocolData, b):
                c = _gen()
                for d in a.data:
                    b(d, c)
                return c

            def __cek(a: DummyProtocolData, f):
                if a.type == 12:
                    c = __gen(a, f)
                elif a.type == 13:
                    c = {}
                    d = a.data
                    for e in d:
                        g = d[e]
                        h = e
                        if isinstance(h, DummyProtocolData):
                            h = __cek(h, f)
                        if isinstance(g, DummyProtocolData):
                            g = __cek(g, f)
                        c[h] = g
                elif a.type in (14, 15):
                    c = []
                    for d in a.data:
                        e = d
                        if isinstance(d, DummyProtocolData):
                            e = __cek(d, f)
                        c.append(e)
                else:
                    c = a.data
                return c

            c = __cek(a, f)
            setattr(b, f"val_{a.id}", c)

        a = _gen()

        def b(c, refs):
            return (
                _genFunc(c, refs, b)
                if type(c.data) in [list, dict]
                else setattr(refs, f"val_{c.id}", c.data)
            )

        if data.data is not None:
            b(data.data, a)
        if self.checkAndGetValue(a, "val_0") is not None:
            return a.val_0
        _ecode = baseException.get("code", 1)
        _emsg = baseException.get("message", 2)
        _emeta = baseException.get("metadata", 3)
        if self.checkAndGetValue(a, "val_1") is not None:
            raise LineServiceException(
                {},
                self.checkAndGetValue(a.val_1, f"val_{_ecode}"),
                self.checkAndGetValue(a.val_1, f"val_{_emsg}"),
                self.checkAndGetValue(a.val_1, f"val_{_emeta}"),
                a.val_1,
            )
        # no vals
        # eg. noop()
        return None


def thrift2dummy(a):
    if type(a) == dict:
        b = {}
        for k, v in a.items():
            b[thrift2dummy(k)] = thrift2dummy(v)
        return b
    elif type(a) == list:
        return [thrift2dummy(a2) for a2 in a]
    elif isinstance(a, DummyProtocolData):
        b = None
        c = []
        if a.type in [2, 3, 4, 6, 8, 10, 11]:
            b = a.data
        elif a.type == 12:
            b = [thrift2dummy(a2) for a2 in a.data]
        elif a.type == 13:
            b = [a.subType[0], a.subType[1], thrift2dummy(a.data)]
        elif a.type in [14, 15]:
            b = [a.subType[0], thrift2dummy(a.data)]
        else:
            raise ValueError(f"Not supported type: {a.type}")
        if a.id is not None:
            return [a.type, a.id, b]
        return b
    else:
        raise ValueError(f"[thrift2dummy] not support `{type(a)}`: {a}")


def doLoopReq(
    req, data, currCount: int = 0, maxRetryCount: int = 5, retryTimeDelay: int = 8
):
    currCount += 1
    doRetry = False
    e = None
    try:
        res = req(**data)
    except httpx.ConnectTimeout as ex:
        doRetry = True
        e = ex
    except (httpx.ReadTimeout, requests.exceptions.ReadTimeout) as ex:
        currCount -= 1
        doRetry = True
        e = ex
    except httpx.ReadError as ex:
        doRetry = True
        e = ex
    except httpx.ConnectError as ex:
        currCount -= 1
        retryTimeDelay += 1
        doRetry = True
        e = ex
    except httpx.RemoteProtocolError as ex:
        currCount -= 1
        doRetry = True
        e = ex
    if doRetry:
        if currCount > maxRetryCount:
            raise e
        time.sleep(retryTimeDelay)
        return doLoopReq(req, data, currCount, maxRetryCount, retryTimeDelay)
    return res



from .client import CHRLINE

# ORIGINAL LICENSE
__copyright__ = "Copyright 2020-2022 by DeachSword"
__version__ = "2.0.0"
__license__ = "BSD-3-Clause"
__author__ = "DeachSword"
__url__ = "http://github.com/DeachSword"

__all__ = ["CHRLINE"]



# -*- coding: utf-8 -*-
"""
Author: YinMo
Version: 2.1.0
Description: support pm and group chat!
"""
import hashlib
import json
import os
import base64
import axolotl_curve25519 as Curve25519
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad

from .exceptions import LineServiceException


class E2EE:
    def getE2EELocalPublicKey(self, mid, keyId):
        toType = self.getToType(mid)
        if toType == 0:
            fd = ".e2eePublicKeys"
            fn = f"key_id_{keyId}.json"
            key = None
            if keyId is not None:
                key = self.getCacheData(fd, fn, False)
            if key is None:
                receiver_key_data = self.negotiateE2EEPublicKey(mid)
                specVersion = self.checkAndGetValue(receiver_key_data, "specVersion", 3)
                if specVersion == -1:
                    raise Exception(f"Not support E2EE on {mid}")
                publicKey = self.checkAndGetValue(receiver_key_data, "publicKey", 2)
                receiverKeyId = self.checkAndGetValue(publicKey, "keyId", 2)
                receiverKeyData = self.checkAndGetValue(publicKey, "keyData", 4)
                if receiverKeyId == keyId:
                    key = base64.b64encode(receiverKeyData)
                    self.saveCacheData(fd, fn, key.decode(), False)
                else:
                    raise Exception(
                        f"E2EE key id-{keyId} not found on {mid}, key id should be {receiverKeyId}"
                    )
        else:
            fd = ".e2eeGroupKeys"
            fn = f"{mid}.json"
            key = self.getCacheData(fd, fn, False)
            if keyId is not None and key is not None:
                keyData = json.loads(key)
                if keyId != keyData["keyId"]:
                    self.log(f"keyId mismatch: {mid}")
                    key = None
            else:
                key = None
            if key is None:
                try:
                    E2EEGroupSharedKey = self.getLastE2EEGroupSharedKey(2, mid)
                except LineServiceException as e:
                    if e.code == 5:
                        self.log(f"E2EE key not registered on {mid}: {e.message}")
                        E2EEGroupSharedKey = self.tryRegisterE2EEGroupKey(mid)
                groupKeyId = self.checkAndGetValue(E2EEGroupSharedKey, "groupKeyId", 2)
                creator = self.checkAndGetValue(E2EEGroupSharedKey, "creator", 3)
                creatorKeyId = self.checkAndGetValue(
                    E2EEGroupSharedKey, "creatorKeyId", 4
                )
                receiver = self.checkAndGetValue(E2EEGroupSharedKey, "receiver", 5)
                receiverKeyId = self.checkAndGetValue(
                    E2EEGroupSharedKey, "receiverKeyId", 6
                )
                encryptedSharedKey = self.checkAndGetValue(
                    E2EEGroupSharedKey, "encryptedSharedKey", 7
                )
                selfKey = base64.b64decode(
                    self.getE2EESelfKeyDataByKeyId(receiverKeyId)["privKey"]
                )
                creatorKey = self.getE2EELocalPublicKey(creator, creatorKeyId)
                aesKey = self.generateSharedSecret(selfKey, creatorKey)
                aes_key = self.getSHA256Sum(aesKey, b"Key")
                aes_iv = self._xor(self.getSHA256Sum(aesKey, b"IV"))
                aes = AES.new(aes_key, AES.MODE_CBC, aes_iv)
                try:
                    decrypted = unpad(aes.decrypt(encryptedSharedKey), 16)
                except ValueError:
                    decrypted = aes.decrypt(encryptedSharedKey)
                self.log(f"[getE2EELocalPublicKey] decrypted: {decrypted}", True)
                data = {
                    "privKey": base64.b64encode(decrypted).decode(),
                    "keyId": groupKeyId,
                }
                key = json.dumps(data)
                self.saveCacheData(fd, fn, key, False)
            return json.loads(key)
        return base64.b64decode(key)

    def generateSharedSecret(self, private_key, public_key):
        return Curve25519.calculateAgreement(bytes(private_key), bytes(public_key))

    def tryRegisterE2EEGroupKey(self, group_mid: str):
        E2EEPublicKeys = self.getLastE2EEPublicKeys(group_mid)
        members = []
        keyIds = []
        encryptedSharedKeys = []
        selfKeyId = [
            self.checkAndGetValue(E2EEPublicKeys[key], "keyId", 2)
            for key in E2EEPublicKeys
            if key == self.mid
        ][0]
        selfKey = base64.b64decode(self.getE2EESelfKeyDataByKeyId(selfKeyId)["privKey"])
        private_key = Curve25519.generatePrivateKey(bytes(32))  # ios patch
        # you can use bytes(32) for LINE Android & PC ver. but it will failed to decrypt on iOS & ChromeOS
        for key_mid in E2EEPublicKeys:
            members.append(key_mid)
            key = E2EEPublicKeys[key_mid]
            keyId = self.checkAndGetValue(key, "keyId", 2)
            keyData = self.checkAndGetValue(key, "keyData", 4)
            keyIds.append(keyId)
            aesKey = self.generateSharedSecret(selfKey, keyData)
            aes_key = self.getSHA256Sum(aesKey, b"Key")
            aes_iv = self._xor(self.getSHA256Sum(aesKey, b"IV"))
            aes = AES.new(aes_key, AES.MODE_CBC, aes_iv)
            encryptedSharedKey = aes.encrypt(pad(private_key, 16))
            encryptedSharedKeys.append(encryptedSharedKey)
        E2EEGroupSharedKey = self.registerE2EEGroupKey(
            1, group_mid, members, keyIds, encryptedSharedKeys
        )
        self.log(f"E2EE key register on {group_mid}: {E2EEGroupSharedKey}")
        return E2EEGroupSharedKey

    def _xor(self, buf):
        buf_length = int(len(buf) / 2)
        buf2 = bytearray(buf_length)
        for i in range(buf_length):
            buf2[i] = buf[i] ^ buf[buf_length + i]
        return bytes(buf2)

    def getSHA256Sum(self, *args):
        instance = hashlib.sha256()
        for arg in args:
            if isinstance(arg, str):
                arg = arg.encode()
            instance.update(arg)
        return instance.digest()

    def _encryptAESECB(self, aes_key, plain_data):
        aes = AES.new(aes_key, AES.MODE_ECB)
        return aes.encrypt(plain_data)

    def decryptKeyChain(self, publicKey, privateKey, encryptedKeyChain):
        shared_secret = self.generateSharedSecret(privateKey, publicKey)
        aes_key = self.getSHA256Sum(shared_secret, "Key")
        aes_iv = self._xor(self.getSHA256Sum(shared_secret, "IV"))
        aes = AES.new(aes_key, AES.MODE_CBC, aes_iv)
        keychain_data = aes.decrypt(encryptedKeyChain)
        key = keychain_data.hex()
        key = bin2bytes(key)
        tc = self.TCompactProtocol(self, passProtocol=True)
        tc.data = key
        key = tc.x(False)[1]
        public_key = bytes(key[0][4])
        private_key = bytes(key[0][5])
        return [private_key, public_key]

    def encryptDeviceSecret(self, publicKey, privateKey, encryptedKeyChain):
        shared_secret = self.generateSharedSecret(privateKey, publicKey)
        aes_key = self.getSHA256Sum(shared_secret, "Key")
        encryptedKeyChain = self._xor(self.getSHA256Sum(encryptedKeyChain))
        keychain_data = self._encryptAESECB(aes_key, encryptedKeyChain)
        return keychain_data

    def generateAAD(self, a, b, c, d, e=2, f=0):
        aad = b""
        aad += a.encode()
        aad += b.encode()
        aad += bytes(self.getIntBytes(c))
        aad += bytes(self.getIntBytes(d))
        aad += bytes(self.getIntBytes(e))  # e2ee version
        aad += bytes(self.getIntBytes(f))  # content type
        return aad

    def encryptE2EEMessage(
        self, to, text, specVersion=2, isCompact=False, contentType=0
    ):
        _from = self.mid
        selfKeyData = self.getE2EESelfKeyData(_from)
        if len(to) == 0 or self.getToType(to) not in [0, 1, 2]:
            raise Exception("Invalid mid")
        senderKeyId = selfKeyData["keyId"]
        if self.getToType(to) == 0:
            private_key = base64.b64decode(selfKeyData["privKey"])
            receiver_key_data = self.negotiateE2EEPublicKey(to)
            specVersion = self.checkAndGetValue(receiver_key_data, "specVersion", 3)
            if specVersion == -1:
                raise Exception(f"Not support E2EE on {to}")
            publicKey = self.checkAndGetValue(receiver_key_data, "publicKey", 2)
            receiverKeyId = self.checkAndGetValue(publicKey, "keyId", 2)
            receiverKeyData = self.checkAndGetValue(publicKey, "keyData", 4)
            keyData = self.generateSharedSecret(bytes(private_key), receiverKeyData)
        else:
            groupK = self.getE2EELocalPublicKey(to, None)
            privK = base64.b64decode(groupK["privKey"])
            pubK = base64.b64decode(selfKeyData["pubKey"])
            receiverKeyId = groupK["keyId"]
            keyData = self.generateSharedSecret(privK, pubK)
        if contentType == 15:
            chunks = self.encryptE2EELocationMessage(
                senderKeyId,
                receiverKeyId,
                keyData,
                specVersion,
                text,
                to,
                _from,
                isCompact=isCompact,
            )
        else:
            chunks = self.encryptE2EETextMessage(
                senderKeyId,
                receiverKeyId,
                keyData,
                specVersion,
                text,
                to,
                _from,
                isCompact=isCompact,
            )
        return chunks

    def encryptE2EETextMessage(
        self,
        senderKeyId,
        receiverKeyId,
        keyData,
        specVersion,
        text,
        to,
        _from,
        isCompact=False,
    ):
        salt = os.urandom(16)
        gcmKey = self.getSHA256Sum(keyData, salt, b"Key")
        aad = self.generateAAD(to, _from, senderKeyId, receiverKeyId, specVersion, 0)
        sign = os.urandom(16)
        data = json.dumps({"text": text}).encode()
        encData = self.encryptE2EEMessageV2(data, gcmKey, sign, aad)
        bSenderKeyId = bytes(self.getIntBytes(senderKeyId))
        bReceiverKeyId = bytes(self.getIntBytes(receiverKeyId))
        if isCompact:
            compact = self.TCompactProtocol(self)
            bSenderKeyId = bytes(compact.writeI32(int(senderKeyId)))
            bReceiverKeyId = bytes(compact.writeI32(int(receiverKeyId)))
        self.log(f"senderKeyId: {senderKeyId} ({bSenderKeyId})", True)
        self.log(f"receiverKeyId: {receiverKeyId} ({bReceiverKeyId})", True)
        return [salt, encData, sign, bSenderKeyId, bReceiverKeyId]

    def encryptE2EELocationMessage(
        self,
        senderKeyId,
        receiverKeyId,
        keyData,
        specVersion,
        location,
        to,
        _from,
        isCompact=False,
    ):
        salt = os.urandom(16)
        gcmKey = self.getSHA256Sum(keyData, salt, b"Key")
        aad = self.generateAAD(to, _from, senderKeyId, receiverKeyId, specVersion, 15)
        sign = os.urandom(16)
        data = json.dumps({"location": location}).encode()
        encData = self.encryptE2EEMessageV2(data, gcmKey, sign, aad)
        bSenderKeyId = bytes(self.getIntBytes(senderKeyId))
        bReceiverKeyId = bytes(self.getIntBytes(receiverKeyId))
        if isCompact:
            compact = self.TCompactProtocol(self)
            bSenderKeyId = bytes(compact.writeI32(int(senderKeyId)))
            bReceiverKeyId = bytes(compact.writeI32(int(receiverKeyId)))
        self.log(f"senderKeyId: {senderKeyId} ({bSenderKeyId})", True)
        self.log(f"receiverKeyId: {receiverKeyId} ({bReceiverKeyId})", True)
        return [salt, encData, sign, bSenderKeyId, bReceiverKeyId]

    def encryptE2EEMessageV2(self, data, gcmKey, nonce, aad):
        aesgcm = AESGCM(gcmKey)
        return aesgcm.encrypt(nonce, data, aad)

    def decryptE2EETextMessage(self, messageObj, isSelf=True):
        _from = self.checkAndGetValue(messageObj, "_from", 1)
        to = self.checkAndGetValue(messageObj, "to", 2)
        toType = self.checkAndGetValue(messageObj, "toType", 3)
        metadata = self.checkAndGetValue(messageObj, "contentMetadata", 18)
        specVersion = metadata.get("e2eeVersion", "2")
        contentType = self.checkAndGetValue(messageObj, "contentType", 15)
        chunks = self.checkAndGetValue(messageObj, "chunks", 20)
        for i in range(len(chunks)):
            if isinstance(chunks[i], str):
                chunks[i] = chunks[i].encode()
        senderKeyId = byte2int(chunks[3])
        receiverKeyId = byte2int(chunks[4])
        self.log(f"senderKeyId: {senderKeyId}", True)
        self.log(f"receiverKeyId: {receiverKeyId}", True)

        selfKey = self.getE2EESelfKeyData(self.mid)
        privK = base64.b64decode(selfKey["privKey"])
        if toType == 0:
            pubK = self.getE2EELocalPublicKey(
                to if isSelf else _from, receiverKeyId if isSelf else senderKeyId
            )
        else:
            groupK = self.getE2EELocalPublicKey(to, receiverKeyId)
            privK = base64.b64decode(groupK["privKey"])
            pubK = base64.b64decode(selfKey["pubKey"])
            if _from != self.mid:
                pubK = self.getE2EELocalPublicKey(_from, senderKeyId)

        if specVersion == "2":
            decrypted = self.decryptE2EEMessageV2(
                to, _from, chunks, privK, pubK, specVersion, contentType
            )
        else:
            decrypted = self.decryptE2EEMessageV1(chunks, privK, pubK)
        return decrypted.get("text", "")

    def decryptE2EELocationMessage(self, messageObj, isSelf=True):
        _from = self.checkAndGetValue(messageObj, "_from", 1)
        to = self.checkAndGetValue(messageObj, "to", 2)
        toType = self.checkAndGetValue(messageObj, "toType", 3)
        metadata = self.checkAndGetValue(messageObj, "contentMetadata", 18)
        specVersion = metadata.get("e2eeVersion", "2")
        contentType = self.checkAndGetValue(messageObj, "contentType", 15)
        chunks = self.checkAndGetValue(messageObj, "chunks", 20)
        for i in range(len(chunks)):
            if isinstance(chunks[i], str):
                chunks[i] = chunks[i].encode()
        senderKeyId = byte2int(chunks[3])
        receiverKeyId = byte2int(chunks[4])
        self.log(f"senderKeyId: {senderKeyId}", True)
        self.log(f"receiverKeyId: {receiverKeyId}", True)

        selfKey = self.getE2EESelfKeyData(self.mid)
        privK = base64.b64decode(selfKey["privKey"])
        if toType == 0:
            pubK = self.getE2EELocalPublicKey(
                to, receiverKeyId if isSelf else senderKeyId
            )
        else:
            groupK = self.getE2EELocalPublicKey(to, receiverKeyId)
            privK = base64.b64decode(groupK["privKey"])
            pubK = base64.b64decode(selfKey["pubKey"])
            if _from != self.mid:
                pubK = self.getE2EELocalPublicKey(_from, senderKeyId)

        if specVersion == "2":
            decrypted = self.decryptE2EEMessageV2(
                to, _from, chunks, privK, pubK, specVersion, contentType
            )
        else:
            decrypted = self.decryptE2EEMessageV1(chunks, privK, pubK)
        return decrypted.get("location", None)

    def decryptE2EEMessageV1(self, chunks, privK, pubK):
        salt = chunks[0]
        message = chunks[1]
        sign = chunks[2]
        aesKey = self.generateSharedSecret(privK, pubK)
        aes_key = self.getSHA256Sum(aesKey, salt, b"Key")
        aes_iv = self._xor(self.getSHA256Sum(aesKey, salt, "IV"))
        aes = AES.new(aes_key, AES.MODE_CBC, aes_iv)
        decrypted = aes.decrypt(message)
        self.log(f"decrypted: {decrypted}", True)
        decrypted = unpad(decrypted, 16)
        return json.loads(decrypted)

    def decryptE2EEMessageV2(
        self, to, _from, chunks, privK, pubK, specVersion=2, contentType=0
    ):
        salt = chunks[0]
        message = chunks[1]
        sign = chunks[2]
        senderKeyId = byte2int(chunks[3])
        receiverKeyId = byte2int(chunks[4])

        aesKey = self.generateSharedSecret(privK, pubK)
        gcmKey = self.getSHA256Sum(aesKey, salt, b"Key")
        aad = self.generateAAD(
            to, _from, senderKeyId, receiverKeyId, specVersion, contentType
        )

        aesgcm = AESGCM(gcmKey)
        decrypted = aesgcm.decrypt(sign, message, aad)
        self.log(f"decrypted: {decrypted}", True)
        return json.loads(decrypted)


def byte2int(t):
    e = 0
    i = 0
    s = len(t)
    for i in range(s):
        e = 256 * e + t[i]
    return e


def bin2bytes(k):
    e = []
    for i in range(int(len(k) / 2)):
        _i = int(k[i * 2 : i * 2 + 2], 16)
        e.append(_i)
    return bytearray(e)



# -*- coding: utf-8 -*-

class OaMembershipService(object):
    OaMembershipService_REQ_TYPE = 4
    OaMembershipService_RES_TYPE = 4
    OaMembershipService_API_PATH = "/EXT/bot/oafan"

    def __init__(self):
        pass

    def activateSubscription(self, uniqueKey: str, activeStatus: str = 1):
        """
        - activeStatus:
            INACTIVE(0),
            ACTIVE(1);
        """
        params = [
            [12, 1, [
                [11, 1, uniqueKey],
                [8, 2, activeStatus]
            ]]
        ]
        sqrd = self.generateDummyProtocol(
            'activateSubscription', params, self.OaMembershipService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.OaMembershipService_API_PATH, sqrd, self.OaMembershipService_RES_TYPE)

    def activateMembership(self, uniqueKey: str, activeStatus: str = 1):
        """
        - activeStatus:
            INACTIVE(0),
            ACTIVE(1);
        """
        params = [
            [12, 1, [
                [11, 1, uniqueKey],
                [8, 2, activeStatus]
            ]]
        ]
        sqrd = self.generateDummyProtocol(
            'activateMembership', params, self.OaMembershipService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.OaMembershipService_API_PATH, sqrd, self.OaMembershipService_RES_TYPE)

    def getJoinedMembership(self, uniqueKey: str):
        params = [
            [12, 1, [
                [11, 1, uniqueKey],
            ]]
        ]
        sqrd = self.generateDummyProtocol(
            'getJoinedMembership', params, self.OaMembershipService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.OaMembershipService_API_PATH, sqrd, self.OaMembershipService_RES_TYPE)

    def getJoinedMemberships(self):
        params = []
        sqrd = self.generateDummyProtocol(
            'getJoinedMemberships', params, self.OaMembershipService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.OaMembershipService_API_PATH, sqrd, self.OaMembershipService_RES_TYPE)

    def getOrderInfo(self, uniqueKey: str):
        params = [
            [12, 1, [
                [11, 1, uniqueKey],
            ]]
        ]
        sqrd = self.generateDummyProtocol(
            'getOrderInfo', params, self.OaMembershipService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.OaMembershipService_API_PATH, sqrd, self.OaMembershipService_RES_TYPE)



# -*- coding: utf-8 -*-

class PrimaryQrCodeMigrationPreparationService(object):
    PQCMPS_REQ_TYPE = 4
    PQCMPS_RES_TYPE = 4
    PQCMPS_API_PATH = "/EXT/auth/feature-user/api/primary/mig/qr/prepare"

    def __init__(self):
        pass
        
    def createQRMigrationSession(self):
        params = [
            [12, 1, []]
        ]
        sqrd = self.generateDummyProtocol(
            'createSession', params, self.PQCMPS_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(
            self.PQCMPS_API_PATH ,sqrd, self.PQCMPS_RES_TYPE)
        
    def sendEncryptedE2EEKey(self, sessionId: str, recoveryKey: str):
        params = [
            [12, 1, [
                [11, 1, sessionId],
                [12, 2, [
                    [11, 1, recoveryKey],
                    [11, 2, backupBlobPayload]
                ]]
            ]]
        ]
        sqrd = self.generateDummyProtocol(
            'sendEncryptedE2EEKey', params, self.PQCMPS_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(
            self.PQCMPS_API_PATH ,sqrd, self.PQCMPS_RES_TYPE)


# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING

from .BaseService import BaseServiceStruct

if TYPE_CHECKING:
    from CHRLINE import CHRLINE


class LoginService(object):
    LoginService_REQ_TYPE = 4  # BASE_VAL
    LoginService_RES_TYPE = 4  # BASE_VAL
    LoginService_API_PATH = None  # BASE_VAL

    def __init__(self):
        pass

    def qrCodeLoginForSecure(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!

        GENERATED BY YinMo0913_DeachSword-DearSakura_v1.0.5.py
        DATETIME: 10/31/2022, 08:17:56
        """
        raise Exception("qrCodeLoginForSecure is not implemented")
        METHOD_NAME = "qrCodeLoginForSecure"
        params = []
        sqrd = self.generateDummyProtocol(
            METHOD_NAME, params, self.LoginService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.LoginService_API_PATH, sqrd, self.LoginService_RES_TYPE
        )

    def qrCodeLoginV2ForSecure(
        self, authSessionId: str, modelName: str, systemName: str, secureCode: str, autoLoginIsRequired: bool = True
    ):
        """
        Qr code login v2 for secure.

        GENERATED BY YinMo0913_DeachSword-DearSakura_v1.0.5.py
        DATETIME: 10/31/2022, 08:17:56
        """
        METHOD_NAME = "qrCodeLoginV2ForSecure"
        params = [
            [11, 1, authSessionId],
            [11, 2, systemName],
            [11, 3, modelName],
            [2, 4, autoLoginIsRequired],
            [11, 5, secureCode],
        ]
        return SecondaryQrCodeLoginServiceStruct.SendRequestByName(
            self, METHOD_NAME, params
        )

    def createQrCodeForSecure(self, qrcode):
        """
        Create qr code for secure.

        GENERATED BY YinMo0913_DeachSword-DearSakura_v1.0.5.py
        DATETIME: 10/31/2022, 08:17:56
        """
        METHOD_NAME = "createQrCodeForSecure"
        params = [[11, 1, qrcode]]
        return SecondaryQrCodeLoginServiceStruct.SendRequestByName(
            self, METHOD_NAME, params
        )


class SecondaryQrCodeLoginServiceStruct(BaseServiceStruct):
    @staticmethod
    def SendRequestByName(client: "CHRLINE", name: str, request: str):
        payload = __class__.BaseRequest(request)
        sqrd = client.generateDummyProtocol(name, payload, 4)
        return client.postPackDataAndGetUnpackRespData(
            "/acct/lgn/sq/v1",
            sqrd,
            4,
            readWith=f"SecondaryQrCodeLoginService.{name}",
        )



# -*- coding: utf-8 -*-
import threading
import traceback


class Poll(object):

    def __init__(self):
        self.subscriptionId = 0
        self.eventSyncToken = None

    def __fetchOps(self, count=100):
        fetchOps = self.fetchOps
        if self.DEVICE_TYPE in self.SYNC_SUPPORT:
            fetchOps = self.sync
        ops = fetchOps(self.revision, count)
        if 'error' in ops:
            raise Exception(ops['error'])
        for op in ops:
            opType = self.checkAndGetValue(op, 'type', 3)
            if opType != -1:
                self.setRevision(self.checkAndGetValue(op, 'revision', 1))
            yield op

    def __fetchMyEvents(self, count: int = 100, initLastSyncToken: bool = False):
        resp = self.fetchMyEvents(self.subscriptionId, self.eventSyncToken, count)
        events = self.checkAndGetValue(resp, 'events', 2)
        for event in events:
            syncToken = self.checkAndGetValue(event, 'syncToken', 5)
            self.setEventSyncToken(syncToken)
            yield event
        if initLastSyncToken:
            syncToken = self.checkAndGetValue(resp, 'syncToken', 3)
            self.setEventSyncToken(syncToken)

    def __execute(self, op, func):
        try:
            func(op, self)
        except Exception as e:
            self.log(traceback.format_exc())

    def setRevision(self, revision):
        if revision is None:
            self.log(f'revision is None!!')
            revision = 0
        self.revision = max(revision, self.revision)

    def setEventSyncToken(self, syncToken):
        if syncToken is None:
            self.log(f'syncToken is None!!')
            syncToken = 0
        if self.eventSyncToken is None:
            self.eventSyncToken = syncToken
        else:
            self.eventSyncToken = max(int(syncToken), int(self.eventSyncToken))

    def trace(self, func, isThreading=True):
        while self.is_login:
            for op in self.__fetchOps():
                opType = self.checkAndGetValue(op, 'type', 'val_3', 3)
                if opType != 0 and opType != -1:
                    if isThreading:
                        _td = threading.Thread(target=self.__execute, args=(op, func))
                        _td.daemon = True
                        _td.start()
                    else:
                        self.__execute(op, func)
